package java.nio.file;

import java.io.IOException;
import sun.nio.fs.BasicFileAttributesHolder;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.Iterator;
import java.util.Collection;
import java.util.ArrayDeque;
import java.io.Closeable;

class FileTreeWalker implements Closeable
{
    private final boolean followLinks;
    private final LinkOption[] linkOptions;
    private final int maxDepth;
    private final ArrayDeque<FileTreeWalker.DirectoryNode> stack;
    private boolean closed;
    
    FileTreeWalker(Collection<FileVisitOption> collection, int maxDepth) {
        this.stack = new ArrayDeque();
        boolean followLinks = false;
        Iterator<Object> iterator = collection.iterator();
        while (iterator.hasNext()) {
            switch ((FileVisitOption)iterator.next()) {
                case FOLLOW_LINKS: {
                    followLinks = true;
                    continue;
                }
                default: {
                    throw new AssertionError("Should not get here");
                }
            }
        }
        if (maxDepth < 0) {
            throw new IllegalArgumentException("'maxDepth' is negative");
        }
        this.followLinks = followLinks;
        this.linkOptions = (followLinks ? new LinkOption[0] : new LinkOption[] { LinkOption.NOFOLLOW_LINKS });
        this.maxDepth = maxDepth;
        return;
    }
    
    private BasicFileAttributes getAttributes(Path path, boolean b) throws IOException {
        if (b && path instanceof BasicFileAttributesHolder && System.getSecurityManager() == null) {
            BasicFileAttributes value = ((BasicFileAttributesHolder)path).get();
            if (value != null && (!this.followLinks || !value.isSymbolicLink())) {
                return value;
            }
        }
        BasicFileAttributes basicFileAttributes;
        try {
            basicFileAttributes = Files.readAttributes(path, BasicFileAttributes.class, this.linkOptions);
        }
        catch (IOException ex) {
            if (!this.followLinks) {
                throw ex;
            }
            basicFileAttributes = Files.readAttributes(path, BasicFileAttributes.class, new LinkOption[] { LinkOption.NOFOLLOW_LINKS });
        }
        return basicFileAttributes;
    }
    
    private boolean wouldLoop(Path path, Object o) {
        for (FileTreeWalker.DirectoryNode directoryNode : this.stack) {
            Object key = directoryNode.key();
            if (o != null && key != null) {
                if (o.equals(key)) {
                    return true;
                }
                continue;
            }
            else {
                try {
                    if (Files.isSameFile(path, directoryNode.directory())) {
                        return true;
                    }
                    continue;
                }
                catch (IOException | SecurityException ex) {}
            }
        }
        return false;
    }
    
    private FileTreeWalker.Event visit(Path path, boolean b, boolean b2) {
        BasicFileAttributes attributes;
        try {
            attributes = this.getAttributes(path, b2);
        }
        catch (IOException ex) {
            return new FileTreeWalker.Event(FileTreeWalker.EventType.ENTRY, path, ex);
        }
        catch (SecurityException ex2) {
            if (b) {
                return null;
            }
            throw ex2;
        }
        if (this.stack.size() >= this.maxDepth || !attributes.isDirectory()) {
            return new FileTreeWalker.Event(FileTreeWalker.EventType.ENTRY, path, attributes);
        }
        if (this.followLinks && this.wouldLoop(path, attributes.fileKey())) {
            return new FileTreeWalker.Event(FileTreeWalker.EventType.ENTRY, path, new FileSystemLoopException(path.toString()));
        }
        DirectoryStream<Path> directoryStream;
        try {
            directoryStream = Files.newDirectoryStream(path);
        }
        catch (IOException ex3) {
            return new FileTreeWalker.Event(FileTreeWalker.EventType.ENTRY, path, ex3);
        }
        catch (SecurityException ex4) {
            if (b) {
                return null;
            }
            throw ex4;
        }
        this.stack.push(new FileTreeWalker.DirectoryNode(path, attributes.fileKey(), directoryStream));
        return new FileTreeWalker.Event(FileTreeWalker.EventType.START_DIRECTORY, path, attributes);
    }
    
    FileTreeWalker.Event walk(Path path) {
        if (this.closed) {
            throw new IllegalStateException("Closed");
        }
        FileTreeWalker.Event visit = this.visit(path, false, false);
        assert visit != null;
        return visit;
    }
    
    FileTreeWalker.Event next() {
        FileTreeWalker.DirectoryNode directoryNode = (FileTreeWalker.DirectoryNode)this.stack.peek();
        if (directoryNode == null) {
            return null;
        }
        FileTreeWalker.Event visit;
        do {
            Path path = null;
            IOException cause = null;
            if (!directoryNode.skipped()) {
                Iterator<Path> iterator = directoryNode.iterator();
                try {
                    if (iterator.hasNext()) {
                        path = (Path)iterator.next();
                    }
                }
                catch (DirectoryIteratorException ex) {
                    cause = ex.getCause();
                }
            }
            if (path == null) {
                try {
                    directoryNode.stream().close();
                }
                catch (IOException exception) {
                    if (cause != null) {
                        cause = exception;
                    }
                    else {
                        cause.addSuppressed(exception);
                    }
                }
                this.stack.pop();
                return new FileTreeWalker.Event(FileTreeWalker.EventType.END_DIRECTORY, directoryNode.directory(), cause);
            }
            visit = this.visit(path, true, true);
        } while (visit == null);
        return visit;
    }
    
    void pop() {
        if (!this.stack.isEmpty()) {
            FileTreeWalker.DirectoryNode directoryNode = (FileTreeWalker.DirectoryNode)this.stack.pop();
            try {
                directoryNode.stream().close();
            }
            catch (IOException ex) {}
        }
    }
    
    void skipRemainingSiblings() {
        if (!this.stack.isEmpty()) {
            ((FileTreeWalker.DirectoryNode)this.stack.peek()).skip();
        }
    }
    
    boolean isOpen() {
        return !this.closed;
    }
    
    @Override
    public void close() {
        if (!this.closed) {
            while (!this.stack.isEmpty()) {
                this.pop();
            }
            this.closed = true;
        }
    }
    
    private static class DirectoryNode
    {
        private final Path dir;
        private final Object key;
        private final DirectoryStream<Path> stream;
        private final Iterator<Path> iterator;
        private boolean skipped;
        
        DirectoryNode(Path dir, Object key, DirectoryStream<Path> stream) {
            this.dir = dir;
            this.key = key;
            this.stream = stream;
            this.iterator = stream.iterator();
        }
        
        Path directory() {
            return this.dir;
        }
        
        Object key() {
            return this.key;
        }
        
        DirectoryStream<Path> stream() {
            return this.stream;
        }
        
        Iterator<Path> iterator() {
            return this.iterator;
        }
        
        void skip() {
            this.skipped = true;
        }
        
        boolean skipped() {
            return this.skipped;
        }
    }
    
    enum EventType
    {
        START_DIRECTORY, 
        END_DIRECTORY, 
        ENTRY;
    }
    
    static class Event
    {
        private final FileTreeWalker.EventType type;
        private final Path file;
        private final BasicFileAttributes attrs;
        private final IOException ioe;
        
        private Event(FileTreeWalker.EventType type, Path file, BasicFileAttributes attrs, IOException ioe) {
            this.type = type;
            this.file = file;
            this.attrs = attrs;
            this.ioe = ioe;
        }
        
        Event(FileTreeWalker.EventType eventType, Path path, BasicFileAttributes basicFileAttributes) {
            this(eventType, path, basicFileAttributes, null);
        }
        
        Event(FileTreeWalker.EventType eventType, Path path, IOException ex) {
            this(eventType, path, null, ex);
        }
        
        FileTreeWalker.EventType type() {
            return this.type;
        }
        
        Path file() {
            return this.file;
        }
        
        BasicFileAttributes attributes() {
            return this.attrs;
        }
        
        IOException ioeException() {
            return this.ioe;
        }
    }
}
